/*
 * GlushkovTraversal.h
 *
 *  Created on: 13. 3. 2014
 *      Author: tomas
 */

#ifndef GLUSHKOVTRAVERSAL_H_
#define GLUSHKOVTRAVERSAL_H_

#include <set>
#include <list>
#include <algorithm>

#include <regexp/RegExp.h>
#include <regexp/RegExpElement.h>
#include <regexp/Alternation.h>
#include <regexp/Concatenation.h>
#include <regexp/Iteration.h>
#include <regexp/RegExpSymbol.h>
#include <regexp/RegExpEmpty.h>
#include <regexp/RegExpEpsilon.h>

#include <AlibException.h>

#include "GlushkovSymbol.h"
#include "GlushkovPair.h"
#include "../../include/macros.h"

namespace conversions
{

/**
 * RegExp tree traversal utils for Glushkov algorithm.
 *
 * Thanks to http://www.sciencedirect.com/science/article/pii/S030439759700296X for better follow() solution.
 */
class GlushkovTraversal
{
public:
    /**
     * @param re RegExp to probe
     * @return all RegExpSymbols whichcan start the word.
     */
    static std::set<GlushkovSymbol> first( regexp::RegExp const& re );

    /**
     * @param re RegExp to probe
     * @return all RegExpSymbols that can terminate the word.
     */
    static std::set<GlushkovSymbol> last( regexp::RegExp const& re );

    /**
     * @param re RegExp to probe
     * @param symbol GlushkovSymbol for which we need the follow()
     * @return all symbols that can follow specific symbol in word
     */
    static std::set<GlushkovSymbol> follow( regexp::RegExp const& re, GlushkovSymbol const& symbol );

    /**
     * @param re RegExp to probe
     * @return symbols of regexp tree in order of they occurence in regexp.
     */
    static std::set<GlushkovSymbol> getSymbols( regexp::RegExp const& re );

private:
    /**
     * @param symbol ptr to symbol
     * @param symbolSet set of gl.symbols
     * @return GlushkovSymbol equivalent for RegExpSymbol pointer
     */
    static GlushkovSymbol const& findSymbol( regexp::RegExpSymbol const * const symbol, const std::set<GlushkovSymbol> & symbolSet );

    /**
     * @return bool true if symbol pointer is in this subtree
     */
    static bool pos( GlushkovSymbol const & symbol, regexp::RegExpElement const * const & node );

    static void getSymbols( regexp::RegExpElement const * const & node, std::set<GlushkovSymbol> & alphabet, int & i );

    static std::set<regexp::RegExpSymbol const *> first( regexp::RegExpElement const * const & node );
    static std::set<regexp::RegExpSymbol const *> first( regexp::Alternation const * const & node );
    static std::set<regexp::RegExpSymbol const *> first( regexp::Concatenation const * const & node );
    static std::set<regexp::RegExpSymbol const *> first( regexp::Iteration const * const & node );
    static std::set<regexp::RegExpSymbol const *> first( regexp::RegExpSymbol const * const & node );
    static std::set<regexp::RegExpSymbol const *> first( regexp::RegExpEmpty const * const & node );
    static std::set<regexp::RegExpSymbol const *> first( regexp::RegExpEpsilon const * const & node );

    static std::set<regexp::RegExpSymbol const *> last( regexp::RegExpElement const * const & node );
    static std::set<regexp::RegExpSymbol const *> last( regexp::Alternation const * const & node );
    static std::set<regexp::RegExpSymbol const *> last( regexp::Concatenation const * const & node );
    static std::set<regexp::RegExpSymbol const *> last( regexp::Iteration const * const & node );
    static std::set<regexp::RegExpSymbol const *> last( regexp::RegExpSymbol const * const & node );
    static std::set<regexp::RegExpSymbol const *> last( regexp::RegExpEmpty const * const & node );
    static std::set<regexp::RegExpSymbol const *> last( regexp::RegExpEpsilon const * const & node );

    static bool pos( regexp::RegExpElement const * const & node, regexp::RegExpSymbol const * const & symbSearch );
    static bool pos( regexp::Alternation const * const & node, regexp::RegExpSymbol const * const & symbSearch );
    static bool pos( regexp::Concatenation const * const & node, regexp::RegExpSymbol const * const & symbSearch );
    static bool pos( regexp::Iteration const * const & node, regexp::RegExpSymbol const * const & symbSearch );
    static bool pos( regexp::RegExpSymbol const * const & node, regexp::RegExpSymbol const * const & symbSearch );
    static bool pos( regexp::RegExpEmpty const * const & node, regexp::RegExpSymbol const * const & symbSearch );
    static bool pos( regexp::RegExpEpsilon const * const & node, regexp::RegExpSymbol const * const & symbSearch );

    static std::set<regexp::RegExpSymbol const *> follow( regexp::RegExpElement const * const & node, regexp::RegExpSymbol const * const & symbFollow );
    static std::set<regexp::RegExpSymbol const *> follow( regexp::Alternation const * const & node, regexp::RegExpSymbol const * const & symbFollow );
    static std::set<regexp::RegExpSymbol const *> follow( regexp::Concatenation const * const & node, regexp::RegExpSymbol const * const & symbFollow );
    static std::set<regexp::RegExpSymbol const *> follow( regexp::Iteration const * const & node, regexp::RegExpSymbol const * const & symbFollow );
    static std::set<regexp::RegExpSymbol const *> follow( regexp::RegExpSymbol const * const & node, regexp::RegExpSymbol const * const & symbFollow );
    static std::set<regexp::RegExpSymbol const *> follow( regexp::RegExpEmpty const * const & node, regexp::RegExpSymbol const * const & symbFollow );
    static std::set<regexp::RegExpSymbol const *> follow( regexp::RegExpEpsilon const * const & node, regexp::RegExpSymbol const * const & symbFollow );
};

} /* namespace conversions */

#endif /* GLUSHKOVTRAVERSAL_H_ */
